////////////////////////////////////////////////////////////////////////
/*! 
 *  \file   TiffIO.cpp
 *  \brief  Plugin to enable TIFF I/O in Qt.
 *  \author JD Gascuel
 * 
 * Copyright JD Gascuel, 2005.
 * 
 * Jean-Dominique.Gascuel@imag.fr
 *
 * This software is a computer program whose purpose is to add TIFF image
 * read/write capabilities to the Qt library, hence to any Qt application.
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 * TiffIO is based on the libtiff 3.7.1, with the following copyright :
 *      Copyright (c) 1988-1997 Sam Leffler
 *      Copyright (c) 1991-1997 Silicon Graphics, Inc.
 * See the libtiff-3.7.1/ subdirectory.
 */
////////////////////////////////////////////////////////////////////////
// File History:
//
// $Id: TiffIO.cpp,v 1.10 2005/02/01 14:19:53 gascuel Exp $
//
// 2004-11-03 : Creation.
//              Support reading of RGBA and RGB bits true color images
//              with up to 16 bits per component (automatically downscaled 
//              to 3x8 bits colors), up to 16 bits gray images (in MINISWHITE 
//              or MINISBLACK format, downscaled to 8-bit grays), 
//              Color palette with up to 8 bits.
//              Support all Paletted/RGB/Gray TIFF encoding scheme (in 
//              particular LZW and ZIP), 
//              except JPEG compression, and LAB or CMYK color spaces.
//
//              Support writting of Qt images with 8-bpp, and 24 and 32-bpp.
//              Always use LZW compression. Automatically switches to MINISBLACK
//              when saving 8-bit all gray image. Use PALETTE mode for 8-bit images,
//              and RGBA mode for 24/32 bit images.
//
// 2004-11-04 : Simplified, and added support for 8-bit LAB and CYMK images.
//              Added support for non-file IO, so .gz images is working.
//              Added support for bigEndians as well as littleEndians file.
//
// 2004-11-22 : FIX : regexp should start with a ^, or wrong image could
//              be declared to be TIFF images.
//
// 2004-05-27 : FIX from Till Oliver Knoll, when calling QtTIFFOpen on a memory
//              stream.
// 2005-01-31 : FIX more compatibility issues on HP-UX and AIX, 
//              thanks to Till Oliver Knoll.

#include "libtiff-3.7.1/tiffio.h"

#include <qimage.h>
#include <qiodevice.h>

////////////////////////////////////////////////////////////////////////
/*
 * TIFF Library Qt-specific Routines.  Adapted from tif_win32.c, 
 * Sam Leffler/SGI 1997, and tif_unix 4/5/95 by
 * Scott Wagner (wagner@itek.com), Itek Graphix, Rochester, NY USA
 */

static tsize_t
QtReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
    QImageIO *io = reinterpret_cast<QImageIO*>(fd);
    Q_ASSERT(io && io->ioDevice());
    if( io->ioDevice()->isReadable() )
        return (tsize_t) io->ioDevice()->readBlock((char *)buf, size);
    else
        return 0;
}

static tsize_t
QtWriteProc(thandle_t fd, tdata_t buf, tsize_t size)
{
    QImageIO *io = reinterpret_cast<QImageIO*>(fd);
    Q_ASSERT(io && io->ioDevice());
    if( io->ioDevice()->isWritable() )
        return (tsize_t) io->ioDevice()->writeBlock((char *)buf, size);
    else
        return 0;
}

static toff_t
QtSizeProc(thandle_t fd)
{
    QImageIO *io = reinterpret_cast<QImageIO*>(fd);
    Q_ASSERT(io && io->ioDevice());

    // Beware : there is a bug in QT, size() re-read what
    // disk says, and do not take into account any buffered
    // write pending...
    io->ioDevice()->flush();
    return (toff_t)io->ioDevice()->size();
}

static toff_t
QtSeekProc(thandle_t fd, toff_t offset, int where)
{
    QImageIO *io = reinterpret_cast<QImageIO*>(fd);
    Q_ASSERT(io && io->ioDevice());

    switch(where) {
	case SEEK_CUR:
        offset += io->ioDevice()->at();
		break;
	case SEEK_END:
        offset = QtSizeProc(fd) - offset;
		break;
    }

    if( io->ioDevice()->at(offset) )
        return offset;
    else
    	return 0;
}

static int
QtCloseProc(thandle_t fd)
{
    QImageIO *io = reinterpret_cast<QImageIO*>(fd);
    Q_ASSERT(io && io->ioDevice());
    io->ioDevice()->close();
	return 0;
}

/*
 * Open a TIFF file descriptor for read/writing.
 * Note that TIFFFdOpen and TIFFOpen recognise the character 'u' in the mode
 * string, which forces the file to be opened unmapped.
 */
TIFF*
QtTIFFOpen(QImageIO *io, char *mode)
{
    char *fileName = (char *)io->fileName().latin1();
    if( ! fileName ) fileName = "memory stream";
	TIFF* tif = TIFFClientOpen(fileName, mode, (thandle_t)io,
			QtReadProc, QtWriteProc,
			QtSeekProc, QtCloseProc, QtSizeProc,
			0,  0);
	return tif;
}

////////////////////////////////////////////////////////////////////////

#define TIFF2QT(c)  qRgba(TIFFGetR(c), TIFFGetG(c), TIFFGetB(c), TIFFGetA(c))

static void readTIFF(QImageIO* image)
{
    // Temporary supress unknown tag errors while reading dir :
    TIFFErrorHandler hndlr = TIFFSetWarningHandler(0);
    TIFF *tif = QtTIFFOpen(image, "rBm");   // LZW bogus in H or L mode...
    TIFFSetWarningHandler(hndlr);

    if( tif )
    {
        uint32 width=0, height=0;
        uint16 bps=0, spp=0, mode=0;

        int success = TIFFGetField(tif, TIFFTAG_IMAGEWIDTH,      &width)
                   && TIFFGetField(tif, TIFFTAG_IMAGELENGTH,     &height)
                   && TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   &bps)
                   && TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp)
                   && TIFFGetField(tif, TIFFTAG_PHOTOMETRIC,     &mode)
                   ;

        // Automatic detection of endianness of the platform :
        int msbFirst; { union {int i; char c[4];} u; u.i=0x11223344; msbFirst= (u.c[0]==0x11); }
        // Should we swab the file ?
        bool badIndians = TIFFIsBigEndian(tif) ^ msbFirst;

        image->setDescription( "TIFF 6.0 image");
        QImage img;

        if( (spp==3) || (spp==4) )
        {
            img.create(width, height, 32);
            if( img.isNull() )
            {
                qWarning("Cannot allocate 32 bit image %dx%d", width, height);
                return;
            }
 
            if( !TIFFReadRGBAImageOriented(tif, width, height, (uint32*)img.bits(), ORIENTATION_TOPLEFT) )
            {
                qWarning("Unsupported pixel format while reading TIFF (spp=%d, bps=%d, mode=%d)",
                    spp, bps, mode);
                success = 0;
            }


            if( success && (TIFF2QT(0x11223344) != 0x11223344) )
                for(uint i=0; i<width; i++)
                    for(uint j=0; j<height; ++j)
                        img.setPixel(i,j, TIFF2QT( img.pixel(i,j) ));
        }
        else if( spp==1 )
        {
            img.create(width, height, 8);
            if( img.isNull() )
            {
                qWarning("Cannot allocate 8-bit image %dx%d", width, height);
                return;
            }

            // Reading and translating color map, if any ?
            if( mode==PHOTOMETRIC_PALETTE )
            {
                int colors = 1<<bps;
                img.setNumColors(colors);
                uint16 *red=0, *green=0, *blue=0;
                if( ! TIFFGetField(tif, TIFFTAG_COLORMAP, &red, &green, &blue) )
                    qWarning("Missing palette data in color-palette TIFF image");
                else
                {
                    Q_CHECK_PTR( red ); Q_CHECK_PTR( green ); Q_CHECK_PTR( blue);
                    if( badIndians )
                    {
                        TIFFSwabArrayOfShort(red,   colors);
                        TIFFSwabArrayOfShort(green, colors);
                        TIFFSwabArrayOfShort(blue,  colors);
                    }
                    for(int c=0; c<colors; ++c)
                        img.setColor(c, qRgb(red[c]>>8,green[c]>>8,blue[c]>>8));
                }
            }
            else if( (mode==PHOTOMETRIC_MINISWHITE) || (mode==PHOTOMETRIC_MINISBLACK) )
            {
                int colors = (bps <8) ? 1<<bps : 256;
                img.setNumColors(colors);
                float s = 255.0f/(colors-1);
                for(int c=0; c<colors; ++c)
                    img.setColor(
                        (mode==PHOTOMETRIC_MINISBLACK) ? c : (colors-1-c),
                        qRgb(s*c, s*c, s*c));
            }
            else
                qWarning("Invalid palette in 8-bit TIFF image (mode=%d)", mode);

            // Reading indexes line by line. Always packed on a byte basis.
            if( bps <= 8 )                                      // 8 bits gray/paletted image ?
                for(uint y=0; success && (y<height); ++y)
                    success &= TIFFReadScanline(tif, img.scanLine(y), y);
            else if( bps<=16 )                                   // 16 bits gray/paletted image ?
            {
                uint16 *wine = (uint16*)new char[TIFFScanlineSize(tif)];
                Q_CHECK_PTR(wine);
                for(uint y=0; success && (y<height); ++y)
                {
                    success &= TIFFReadScanline(tif, wine, y);  // Already swabed, if needed.
                    for(uint x=0; x<width; ++x)
                        img.setPixel(x, y, wine[x]>>8);
                }
                delete[] wine;
            }
            else
                qWarning("Unsupported pixel format while reading TIFF (spp=%d, bps=%d, mode=%d)", spp, bps, mode);
        }
        else
            qWarning("Unsupported pixel format while reading TIFF (spp=%d, bps=%d, mode=%d)", spp, bps, mode);

        TIFFClose(tif);

        image->setImage( img );
        image->setStatus(success ? 0 : -2);    // Everything went OK ?
    }
    else
        image->setStatus(-1);   // Error reading image...
}

#define QT2TIFF(c)  (qRed(c) | qGreen(c)<<8 | qBlue(c)<<16 | qAlpha(c)<<24)

static void writeTIFF(QImageIO* image)
{
    TIFF *tif = QtTIFFOpen(image, "wBm");   // LZW bogus in H or L mode...

    if( tif )
    {
        int success = TIFFSetField(tif, TIFFTAG_IMAGEWIDTH,      image->image().width())
                   && TIFFSetField(tif, TIFFTAG_IMAGELENGTH,     image->image().height())
                   && TIFFSetField(tif, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_CONTIG)
                   && TIFFSetField(tif, TIFFTAG_ORIENTATION,     ORIENTATION_TOPLEFT)
                   && TIFFSetField(tif, TIFFTAG_COMPRESSION,     COMPRESSION_LZW)
                   ;

        if( image->image().depth() > 8 )
        {
            QImage img = image->image().convertDepth(32);

            success &= TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,   8)
                    && TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 4)
                    && TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,     PHOTOMETRIC_RGB)
                    ;

            // Need RGB transcoding ?
            if( QT2TIFF(0x11223344) != 0x11223344 )
            {
                Q_UINT32* buffer = new Q_UINT32[img.width()];
                Q_CHECK_PTR(buffer);

                for(int y=0; success && (y<img.height()); ++y)
                {
                    Q_UINT32*p = (Q_UINT32*)img.scanLine(y);
                    Q_CHECK_PTR(p);

                    for(int x=0; x<img.width(); ++x, ++p)
                        buffer[x] = QT2TIFF(p[0]);

                    success &= TIFFWriteScanline(tif, buffer, y);
                }
                delete[] buffer;
            }
            else
                // Write all lines, in order (planar does not support anything else ?)
                for(int y=0; success && (y<img.height()); ++y)
                    success &= TIFFWriteScanline(tif, img.scanLine(y), y);

            image->setStatus( success ? 0 : -2 );
        }
        else if( image->image().depth()==8 )
        {
            const QImage& img = image->image();

            success &= TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,   8)
                    && TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1)
                    ;

            if( img.isGrayscale() )
            {
                success &= TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
            }
            else
            {
                int colors = img.numColors();
                uint16 *red   = new uint16[colors]; Q_CHECK_PTR(red);
                uint16 *green = new uint16[colors]; Q_CHECK_PTR(green);
                uint16 *blue  = new uint16[colors]; Q_CHECK_PTR(blue);
                float scale=65535.0f/255.0f;
                for(int c=0; c<colors; ++c)
                {
                    QRgb rgb = img.color(c);
                    red[c]   = qRed(rgb)  *scale;
                    green[c] = qGreen(rgb)*scale;
                    blue[c]  = qBlue(rgb) *scale;
                }
                success &= TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
                success &= TIFFSetField(tif, TIFFTAG_COLORMAP,    red, green, blue);
                delete[] blue;
                delete [] green;
                delete[] red;
            }

            // Write all lines, in order (planar does not support anything else ?)
            for(int y=0; success && (y<img.height()); ++y)
                success &= TIFFWriteScanline(tif, img.scanLine(y), y);

            image->setStatus( success ? 0 : -2 );
        }

        TIFFFlush(tif);
        TIFFClose(tif);

    }
    else
        image->setStatus(-1);
}

///////////////////////////////////////////////////////////////////////
//                                                                   //
//  Glue needed to install the read/write handlers.                  //
//                                                                   //
///////////////////////////////////////////////////////////////////////

#include <qregexp.h>
#include <qstringlist.h>
#include <qimageformatplugin.h>

class TIFFIOPlugin : public QImageFormatPlugin
{
public:
    virtual QStringList keys() const
    {
        QStringList list;
            list += "TIF";
            list += "TIFF";
        return  list;
    }
    virtual bool installIOHandler(const QString& format)
    {
        if( (format=="TIF") || (format=="TIFF") )
        {
            QImageIO::defineIOHandler(format,
                    "^"                         // Start at header begining!
                    "(MM\\x01[\\x2a\\x2b])"     // TIFF, or Big TIFF, in
                    "|(II[\\x2a\\x2b]\\x01)",   // Big or little endian modes.
                    0,
                    readTIFF,
                    writeTIFF);
            return true;
        }
        else
            return false;
    }
};

Q_EXPORT_PLUGIN(TIFFIOPlugin);